#!/usr/bin/env python
"""
Cuckoo search.

Usage:
  cs.py [options]

Options:
  -h, --help  Show this message and exit.
  -n N        Number of generations in CS. [default: 20]
  --print-level LEVEL
              Print verbose level. [default: 1]
"""
import os
import sys
from docopt import docopt
import numpy as np
from numpy import exp, sin, cos
import random
import copy
from multiprocessing import Process, Pool
from time import time
from scipy.special import gamma

__author__ = "RYO KOBAYASHI"
__version__ = "221223"

_fname_gen = 'out.cs.generations'
_fname_ind = 'out.cs.individuals'

def test_func(var, vranges, **kwargs):
    x,y= var
    res= x**2 +y**2 +100.0*exp(-x**2 -y**2)*sin(2.0*(x+y))*cos(2*(x-y)) \
         +80.0*exp(-(x-1)**2 -(y-1)**2)*cos(x+4*y)*sin(2*x-y) \
         +200.0*sin(x+y)*exp(-(x-3)**2-(y-1)**2)
    return res

def test_write_func(vs,vrs,fname,**kwargs):
    with open(fname,'w') as f:
        for i,v in enumerate(vs):
            vr = vrs[i]
            f.write(' {0:10.3f}  {1:10.3f}  {2:10.3f}\n'.format(v,*vr))
    return None

def wrap(vs,vrs):
    vsnew = copy.copy(vs)
    for i,v in enumerate(vsnew):
        vmin, vmax = vrs[i]
        vsnew[i] = min(max(v,vmin),vmax)
    return vsnew

def update_vrange(vrs,vrsh,all_indivisuals,ntops=100):
    """
    Update variable ranges adaptively using all the individuals information.
    """
    #...Extract top NTOPS individuals from all
    tops = []
    # print('len(all_indivisuals)=',len(all_indivisuals))
    for i,ind in enumerate(all_indivisuals):
        if len(tops) < ntops:  # add the individual
            # print(' i (< ntops)=',i)
            for it,t in enumerate(tops):
                if ind.val < t.val:
                    tops.insert(it,ind)
                    break
            if not ind in tops:
                tops.append(ind)
        else: # insert the individual and pop out the worst one
            # print(' i (>=ntops)=',i)
            for it,t in enumerate(tops):
                if ind.val < t.val:
                    tops.insert(it,ind)
                    break
            if len(tops) > ntops:
                del tops[ntops:len(tops)]

    # print('len(tops)=',len(tops))
    # print('iids= ',[t.iid for t in tops])
    #...Get new ranges
    new_vrs = np.array(vrs)
    vss = np.zeros((len(tops),len(vrs)))
    for i,ind in enumerate(tops):
        vi = ind.vector
        vss[i,:] = vi[:]
        # print('i,vi=',i,vi)
    for j in range(len(new_vrs)):
        # print('j,min,max=',i,min(vss[:,j]),max(vss[:,j]))
        new_vrs[j,0] = max(new_vrs[j,0],min(vss[:,j]))
        new_vrs[j,1] = min(new_vrs[j,1],max(vss[:,j]))

    #...Set best variables center in the ranges
    fbest = tops[0].val
    vbest = tops[0].vector
    for j in range(len(vbest)):
        vjmin = new_vrs[j,0]
        vjmax = new_vrs[j,1]
        wmax = max(abs(vjmin-vbest[j]),abs(vjmax-vbest[j]))
        new_vrs[j,0] = max(min(vjmin,vbest[j]-wmax),vrsh[j,0])
        new_vrs[j,1] = min(max(vjmax,vbest[j]+wmax),vrsh[j,1])
    
    return new_vrs

class Individual:
    """
    Individual class that consists of variables as vector elements.
    """
    def __init__(self, iid, ndim, vranges, loss_func):
        self.iid = iid
        self.ndim = ndim
        self.loss_func = loss_func
        self.vector = np.zeros(self.ndim)
        self.vranges = vranges
        self.val = None

    def set_variable(self,variables):
        if len(variables) != len(self.vector):
            raise ValueError()

        self.vector = variables
        # print('iid, v before wrap,vrs =',self.iid,self.vector,self.vranges)
        self.wrap_range()
        # print('iid, v after wrap,vrs  =',self.iid,self.vector,self.vranges)
        self.val = None
        return None

    def init_random(self):
        for i in range(self.ndim):
            vmin, vmax = self.vranges[i]
            # vmin = self.vranges[i,0]
            # vmax = self.vranges[i,1]
            v = np.random.random()*(vmax -vmin) +vmin
            self.vector[i] = v
            # print(' i,vmin,vmax,v=',i,vmin,vmax,v)
        self.wrap_range()
        self.val = None
        return None

    def wrap_range(self):
        self.vector = wrap(self.vector, self.vranges)

    def calc_loss_func(self,kwargs):
        """
        Compute loss function value using self.loss_func function given in the constructor.
        """
        # print('type(kwargs)=',type(kwargs))
        val = self.loss_func(self.vector, **kwargs)
        # print(' iid,v,val=',self.iid,self.vector,val)
        return val,kwargs['index']

class CS:
    """
    Cuckoo search class.
    """

    def __init__(self, N, F, variables, vranges, vhardlimit, loss_func, write_func,
                 nproc=0,seed=42,**kwargs):
        """
        Conctructor of CS class.

        N:  Number of individuals.
        F:  Fraction of worse individuals to be abondoned.
        loss_func:
            Loss function to be minimized with variables and **kwargs.
        nproc:
            Number of processes used to run N individuals.
        """
        if N < 2:
            raise ValueError('N must be greater than 1 in CS!')
        np.random.seed(seed)
        random.seed(seed)
        self.N = N   # Number of individuals in a generation
        self.F = F   # Fraction of worse individuals to be abondoned
        self.nproc = nproc
        self.ndim = len(variables)
        self.vs = variables
        self.vrs0 = vranges
        self.vrs = copy.copy(self.vrs0)
        self.vrsh = vhardlimit
        self.vws = np.zeros(self.ndim)
        for i in range(self.ndim):
            self.vws[i] = max(self.vrs[i,1] -self.vrs[i,0], 0.0)
        # print('original variables=',self.vs,self.vrs)
        self.loss_func = loss_func
        self.write_func = write_func
        self.kwargs = kwargs
        self.bestind = None
        self.print_level = 0
        if 'print_level' in kwargs.keys():
            self.print_level = int(kwargs['print_level'])
        if 'update_vrange' in kwargs.keys():
            self.update_vrs_per = kwargs['update_vrange']

        self.beta = 1.5
        self.betai = 1.0 /self.beta
        self.usgm = (gamma(1+self.beta)*np.sin(np.pi*self.beta/2)/ \
                     gamma((1+self.beta)/2)*self.beta*2.0**((self.beta-1)/2))**self.betai
        self.vsgm = 1.0

        #...initialize population
        self.population = []
        self.all_indivisuals = []
        self.iidmax = 0
        for i in range(N):
            self.iidmax += 1
            ind = Individual(self.iidmax, self.ndim, self.vrs, self.loss_func)
            if i == 0:
                ind.set_variable(self.vs)
            else:
                ind.init_random()
            self.population.append(ind)

        #...Evaluate loss function values
        prcs = []
        if self.nproc > 0 :  # use specified number of cores by nproc
            pool = Pool(processes=self.nproc)
        else:
            pool = Pool()
            
        for ip,pi in enumerate(self.population):
            kwtmp = copy.copy(self.kwargs)
            kwtmp['index'] = ip
            kwtmp['iid'] = pi.iid
            #prcs.append(pool.apply_async(pi.calc_loss_func, (kwtmp,qs[ip])))
            prcs.append(pool.apply_async(pi.calc_loss_func, (kwtmp,)))
        results = [ res.get() for res in prcs ]
        for res in results:
            val,ip = res
            self.population[ip].val = val

        pool.close()
        
        self.keep_best()
        self.all_indivisuals.extend(self.population)
        if self.print_level > 2:
            for pi in self.population:
                fname = 'in.vars.optzer.{0:d}'.format(pi.iid)
                self.write_variables(pi,
                                     fname=fname,
                                     **self.kwargs)
        else:
            fname = 'in.vars.optzer.{0:d}'.format(self.bestind.iid)
            self.write_variables(self.bestind,
                                 fname=fname,
                                 **self.kwargs)
        return None

    def keep_best(self):
        vals = []
        for i,pi in enumerate(self.population):
            # print('i,val,vec=',i,pi.val,pi.vector)
            if pi.val == None:
                raise ValueError('Something went wrong.')
            vals.append(pi.val)

        minval = min(vals)
        if self.bestind == None or minval < self.bestind.val:
            idx = vals.index(minval)
            self.bestind = copy.deepcopy(self.population[idx])
        return None

    def sort_individuals(self):
        """
        Sort individuals in the population in the ascending order.
        """
        jtop = self.N
        for i in range(self.N):
            jtop -= 1
            for j in range(jtop):
                pj = self.population[j]
                pjp = self.population[j+1]
                if pj.val > pjp.val:
                    self.population[j] = pjp
                    self.population[j+1] = pj
        return None
        
    def run(self,maxiter=100):
        """
        Perfom CS.
        """

        if 'start' in self.kwargs.keys():
            start = self.kwargs['start']
        else:
            start = time()
        fgen = open(_fname_gen,'w')
        find = open(_fname_ind,'w')
        #...Headers
        fgen.write('# {0:>4s}  {1:>8s}  {2:12s}\n'.format('gen','iid','loss'))
        find.write('# {0:>7s}  {1:>12s}'.format('iid', 'loss'))
        for i in range(len(self.population[0].vector)):
            find.write(' {0:>8d}-th'.format(i+1))
        find.write('\n')
        
        for i,ind in enumerate(self.population):
            fgen.write('     0  {0:8d}  {1:12.4e}\n'.format(ind.iid, ind.val))
            find.write(' {0:8d}  {1:12.4e}'.format(ind.iid, ind.val))
            for j,vj in enumerate(ind.vector):
                find.write(' {0:11.4e}'.format(vj))
            find.write('\n')

        if self.print_level > 0:
            print(' step,time,best_iid,best_loss,vars='
                  +' {0:6d} {1:8.1f} {2:5d} {3:8.4f}'.format(0,time()-start,
                                                             self.bestind.iid,
                                                             self.bestind.val),end="")
            for i in range(min(16,self.ndim)):
                print(' {0:6.3f}'.format(self.bestind.vector[i]),end="")
            print('', flush=True)

        #...Create pool before going into maxiter-loop,
        #...since creating pool inside could cause "Too many files" error.
        if self.nproc > 0 :  # use specified number of cores by nproc
            pool = Pool(processes=self.nproc)
        else:
            pool = Pool()
            
        for it in range(maxiter):
            self.sort_individuals()
            #...Create candidates from current population using Levy flight
            candidates = []
            vbest = self.bestind.vector
            for ip,pi in enumerate(self.population):
                vi = pi.vector
                vnew =np.array(vi)
                for iv in range(self.ndim):
                    u = np.random.normal()*self.usgm
                    v = abs(np.random.normal()*self.vsgm)
                    v = max(v,1.0e-8)
                    w = u/v**self.betai
                    zeta = self.vws[iv] *0.01 *w
                    # zeta = self.vws[iv]*0.01 *w *(vi[iv] -vbest[iv])
                    # if ip == 0:
                    #     zeta = self.vws[iv] *0.001 *w
                    # else:
                    #     zeta = 0.01 *w *(vi[iv] -vbest[iv])
                    vnew[iv] = vnew[iv] +zeta*np.random.normal()
                #...create new individual for trial
                # print('ip,vi,vnew=',ip,vi,vnew)
                self.iidmax += 1
                newind = Individual(self.iidmax, self.ndim, self.vrs, self.loss_func)
                newind.set_variable(vnew)
                candidates.append(newind)

            #...Create new completely random candidates
            iab = int((1.0 -self.F)*self.N)
            rnd_candidates = []
            for iv in range(iab,self.N):
                self.iidmax += 1
                newind = Individual(self.iidmax, self.ndim, self.vrs, self.loss_func)
                newind.init_random()
                rnd_candidates.append(newind)

            #...Evaluate loss function values of updated candidates and new random ones
            prcs = []
            for ic,ci in enumerate(candidates):
                kwtmp = copy.copy(self.kwargs)
                kwtmp['index'] = ic
                kwtmp['iid'] = ci.iid
                # prcs.append(Process(target=ci.calc_loss_func, args=(kwtmp,qs[ic])))
                prcs.append(pool.apply_async(ci.calc_loss_func, (kwtmp,)))
            rnd_prcs = []
            for ic,ci in enumerate(rnd_candidates):
                kwtmp = copy.copy(self.kwargs)
                kwtmp['index'] = len(candidates) +ic
                kwtmp['iid'] = ci.iid
                # prcs.append(Process(target=ci.calc_loss_func, args=(kwtmp,qs[ic])))
                rnd_prcs.append(pool.apply_async(ci.calc_loss_func, (kwtmp,)))
            
            results = [ res.get() for res in prcs ]
            rnd_results = [ res.get() for res in rnd_prcs ]

            for res in results:
                val,ic = res
                candidates[ic].val = val
            self.all_indivisuals.extend(candidates)

            for res in rnd_results:
                val,ic_rnd = res
                ic = ic_rnd -len(candidates)
                rnd_candidates[ic].val = val
            self.all_indivisuals.extend(rnd_candidates)

            #...Pick j that is to be compared with i
            js = random.sample(range(self.N),k=self.N)
            #...Decide whether or not to adopt new one
            for jc,jv in enumerate(js):
                pj = self.population[jv]
                cj = candidates[jc]
                dval = cj.val -pj.val
                if dval < 0.0:  # replace with new individual
                    self.population[jv] = cj
                    find.write(' {0:8d}  {1:12.4e}'.format(cj.iid, cj.val))
                    for k,vk in enumerate(cj.vector):
                        find.write(' {0:11.4e}'.format(vk))
                    find.write('\n')
                    find.flush()
                else:
                    pass

            #...Rank individuals
            self.sort_individuals()
            
            #...Replace to-be-abandoned ones with new random ones
            ic = 0
            for iv in range(iab,self.N):
                ci = rnd_candidates[ic]
                ic += 1
                self.population[iv] = ci

            #...Check best
            best_updated = False
            for ic,ci in enumerate(self.population):
                if ci.val < self.bestind.val:
                    self.bestind = ci
                    best_updated = True
            if best_updated:
                fname = 'in.vars.optzer.{0:d}'.format(self.bestind.iid)
                self.write_variables(self.bestind,
                                     fname=fname,
                                     **self.kwargs)
                os.system('cp -f {0:s} in.vars.optzer.best'.format(fname))

            #...Update variable ranges if needed
            if self.update_vrs_per > 0 and (it+1) % self.update_vrs_per == 0:
                self.vrs = update_vrange(self.vrs,self.vrsh,self.all_indivisuals)
                print(' Update variable ranges')
                for i in range(len(self.vrs)):
                    print(' {0:2d}:  {1:7.3f}  {2:7.3f}'.format(i+1,self.vrs[i,0],self.vrs[i,1]))
                #...Set variable ranges of all individuals in the population
                for iv in range(len(self.population)):
                    self.population[iv].vranges = self.vrs
            
            if self.print_level > 0:
                print(' step,time,best_iid,best_loss,vars='
                      +' {0:6d} {1:8.1f} {2:5d} {3:8.4f}'.format(it+1,time()-start,
                                                                 self.bestind.iid,
                                                                 self.bestind.val),end="")
                # print(' step,time,best,vars= {0:6d} {1:8.1f}  {2:8.4f}'.format(it+1, time()-start,
                #                                                                self.bestind.val),end="")
                for i in range(min(16,self.ndim)):
                    print(' {0:6.3f}'.format(self.bestind.vector[i]),end="")
                print('', flush=True)

            for i,ind in enumerate(self.population):
                fgen.write(' {0:5d}  {1:8d}  {2:12.4e}\n'.format(it+1, ind.iid, ind.val))
                fgen.flush()
        fgen.close()
        find.close()
        pool.close()
        #...Finaly write out the best one
        self.write_variables(self.bestind,fname='in.vars.optzer.best',**self.kwargs)
        return None

    def write_variables(self,ind,fname='in.vars.optzer',**kwargs):
        
        vs = ind.vector
        vrs = ind.vranges
        self.write_func(vs,vrs,fname,**kwargs)
        return None


if __name__ == "__main__":

    args = docopt(__doc__)
    n = int(args['-n'])
    kwargs = {}
    kwargs['print_level'] = int(args['--print-level'])


    vs = np.array([1.0, -0.5])
    vrs = np.array([[-1.0, 2.0],[-1.0, 1.0]])
    
    cs = CS(10, 0.25, vs, vrs, test_func, test_write_func, **kwargs)
    cs.run(n)
    



#!/usr/bin/env python
"""
Convert optzer parameters `in.vars.optzer` to parameter files.

Usage:
  opt2prms.py [options] <v-file> <p-files> [<p-files>...]

Options:
  -h, --help  Show this message and exit.
"""
import os
from docopt import docopt

from .io import read_vars_optzer

__author__ = "RYO KOBAYASHI"
__version__ = "221223"

def read_in_fitpot2(infname='in.fitpot'):
    """
    Get specorder, pairs, triplets from in.fitpot.
    """
    if not os.path.exists(infname):
        raise FileNotFoundError(infname)
    
    with open(infname,'r') as f:
        lines = f.readlines()

    mode = None
    specorder = []
    interact = []
    param_files = []
    for line in lines:
        data = line.split()
        if len(data) == 0:
            mode = None
            continue
        if data[0] in ('#','!'):
            mode = None
            continue
        elif data[0] == 'specorder':
            specorder = [ x for x in data[1:] ]
            continue
        elif data[0] == 'interactions':
            num_interact = int(data[1])
            mode = 'interactions'
            continue
        elif data[0] == 'param_files':
            param_files = [ name for name in data[1:] ]
            mode = None
        else:
            if mode == 'interactions':
                if len(data) not in (2,3):
                    raise Exception('len(data) is not 2 nor 3.')
                interact.append(data)
                if len(interact) == num_interact:
                    mode = None
            else:
                mode = None

    return specorder, interact, param_files
    

def read_vars_fitpot(fname='in.vars.fitpot'):
    """
    Read in.vars.fitpot and return data.
    """
    with open(fname,'r') as f:
        lines = f.readlines()

    fpvars = []
    vranges = []
    il = -1
    nv = -1
    while True:
        il += 1
        line = lines[il]
        if line[0] in ('!','#'):  # skip comment line
            il += 1
            continue
        data = line.split()
        if nv < 0:
            nv = int(data[0])
            rc = float(data[1])
            rc3= float(data[2])
        else:
            fpvars.append(float(data[0]))
            vranges.append([ float(x) for x in data[1:3]])
            if len(fpvars) == nv:
                break
    varsfp = {}
    varsfp['rc2'] = rc
    varsfp['rc3'] = rc3
    varsfp['variables'] = fpvars
    varsfp['vranges'] = vranges
    return varsfp
    
def vars2params(vs,**kwargs):
    """
    Conversion from optzer-vars to files specified in param_files in in.optzer.
    The param_files should contain key-phrases such as '{p[0]}' that are converted from fp-vars,
    and the indices in the key-phrases must correspond to those in optzer-vars..
    """
    import re
    try:
        param_fnames = kwargs['param_files']
    except:
        raise
    if type(param_fnames) != list or len(param_fnames) < 1:
        raise ValueError('param_files may not be specified correctly...')
        
    for fname in param_fnames:
        try:
            fcontents = kwargs[fname]
            #...If the format is '{0:.1f}'-style, replace them to '{p[0]:.2f}'-style
            res = re.search(r'\{[0-9]+:',fcontents)
            if res != None:
                fcontents = re.sub(r'\{([0-9]+):',r'{p[\1]:',fcontents)
            new_contents = fcontents.format(p=vs)
        except:
            print('ERROR: Failed to replace the parameters in param_files !!!')
            print(fcontents)
            raise
        newfname = os.path.basename(fname)
        if os.path.exists(newfname):
            raise ValueError(f'{newfname} already exists here !!!\n'
                             +f'Remove {newfname} and try again.')
        with open(newfname,'w') as f:
            f.write(new_contents)
    return None

def main():
    import os,sys
    args = docopt(__doc__.format(os.path.basename(sys.argv[0])))
    # print(args)
    vfile = args['<v-file>']
    pfiles = args['<p-files>']

    kwargs = {}
    kwargs['param_files'] = pfiles
    
    vs,vrs,vrsh,options,vopts = read_vars_optzer(vfile)

    for pfile in pfiles:
        with open(pfile,'r') as f:
            kwargs[pfile] = f.read()
        os.system(f'cp -f {pfile} {pfile}.bak')
    vars2params(vs, **kwargs)
    print(' Convert the following files by replacing with optimized parameters:')
    for pfile in pfiles:
        newpfile = os.path.basename(pfile)
        print(f'   - {pfile} ==> {newpfile}')
    print('')
    return None

if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Optimize parameters of any external program to any target property that can be computed with the programs.

Usage:
  {0:s} [options]

Options:
  -h, --help  Show this message and exit.
  --nproc NPROC
              Number of processes to be used. If it's less than 1, use as many processes as possible. [default: 0]
  --subdir-prefix PREFIX_DIR
              Prefix for subjob directory. [default: subdir_]
  --subjob-script SCRIPT
              Name of script that performs MD and post-processing. [default: subjob.sh]
  --subjob-prefix PREFIX_JOB
              Prefix for performing subjob. [default: ]
  --subjob-timeout TIMEOUT
              Timeout for a subjob in sec. [default: 3600]
  --random-seed SEED
              Random seed for reproducibility, if negative current time in second is applied. [default: -1]
"""

__all__ = ['io','get_best','opt2md','opt2prms']

import os
import sys
import shutil
import glob
from docopt import docopt
import numpy as np
from numpy import sin,cos,sqrt
import subprocess
import time
from datetime import datetime

from .opt2prms import vars2params
from .io import read_in_optzer, write_info, write_vars_optzer, read_vars_optzer, \
    read_data
from .de import DE
from .cs import CS
from .tpe import TPE

__author__ = "RYO KOBAYASHI"
__version__ = "0.1.0"

_infname = 'in.optzer'

def get_data(basedir,prefix=None,**kwargs):
    """
    New implementation of get_data, which loads data to be used to fit parameters.
    If the prefix ref, some special treatment will be done.
    """

    targets = kwargs['target']

    if prefix == 'ref':
        print(' Reference data and weights:')

    data = {}
    for t in targets:
        if prefix == 'ref':
            fname = basedir+'/data.{0:s}.{1:s}'.format(prefix,t)
        else:  # not known about prefix
            files = glob.glob(basedir+f'/data.*.{t:s}')
            fname = files[0]  # assume 1st one as the generated data, which could be wrong
        # print('m,fname=',m,fname)
        try:
            data[t] = read_data(fname,)
        except:
            data[t] = None
            pass
        if prefix == 'ref':
            print('   {0:s}  {1:.3f}'.format(t,data[t]['wdat']))
    return data

def loss_func(tdata,eps=1.0e-8,**kwargs):
    """
    Compute loss function value using general get_data func.
    """
    refdata = kwargs['refdata']
    losses = {}
    L = 0.0
    misval = kwargs['missing_value']
    luplim = kwargs['fval_upper_limit']
    for name in refdata.keys():
        ref = refdata[name]
        wgt = ref['wdat']
        dtype = ref['datatype']
        eps = ref['eps']
        trial = tdata[name]
        if trial == None:
            losses[name] = misval
            L += losses[name] *wgt
            continue
        num = ref['ndat']
        refd = ref['data']
        td = trial['data']
        z2 = 0.0
        sumdiff2 = 0.0
        if dtype[:5] == 'indep':  # independent data
            epss = ref['epss']
            for n in range(num):
                diff = td[n] - refd[n]
                sumdiff2 += diff*diff /epss[n]**2
            if num > 0:
                sumdiff2 /= num
            losses[name] = min(sumdiff2, luplim)
        elif dtype[:3] == 'sep':  # data treated separately
            for n in range(num):
                # print('n=',n)
                diff = td[n] -refd[n]
                sumdiff2 += diff*diff /(refd[n]**2+eps)
            losses[name] = min(sumdiff2, luplim)
        else:  # data treated all together (default)
            for n in range(num):
                # print('n=',n)
                diff = td[n] -refd[n]
                sumdiff2 += diff*diff
                z2 += refd[n]*refd[n]
            losses[name] = min(sumdiff2 /(z2+eps), luplim)

        if 'subject_to' in ref.keys():
            import re
            for prange in ref['subject_to']:
                pid = prange['pid']  # int
                lower0 = prange['lower']  # str
                upper0 = prange['upper']  # str
                penalty = prange['penalty']  # float
                lower1 = re.sub(r'\{([0-9]+)\}',r'{p[\1]}',lower0)
                upper1 = re.sub(r'\{([0-9]+)\}',r'{p[\1]}',upper0)
                lower2 = lower1.format(p=td)
                upper2 = upper1.format(p=td)
                lower = eval(lower2)
                upper = eval(upper2)
                if lower > upper:
                    raise ValueError('lower is greater than upper in subject_to ',pid)
                if not (lower < td[pid] < upper):
                    losses[name] += penalty

        L += losses[name] *wgt
        
    if kwargs['print_level'] > 0:
        print('   iid,losses= {0:8d}'.format(kwargs['iid']),end='')
        for k in losses.keys():
            loss = losses[k]
            print(' {0:10.4f}'.format(loss),end='')
        print(' {0:11.5f}'.format(L),flush=True)
    return L

def func_wrapper(variables, **kwargs):
    """
    Wrapper function for the above loss_func().
    This converts variables to be optimized to parameters for the external program,
    perform it, then get the target properties.
    Then pass them to the the loss_func().
    """
    refdata = kwargs['refdata']
    wgts = kwargs['weights']
    refdata = kwargs['refdata']
    subjobscript = kwargs['subjob-script']
    subdir = kwargs['subdir-prefix'] +'{0:03d}'.format(kwargs['index'])
    print_level = kwargs['print_level']

    #...Create param_files in each subdir
    varsfp = {}
    varsfp['variables'] = variables
    cwd = os.getcwd()
    if not os.path.exists(subdir):
        os.mkdir(subdir)
        shutil.copy(subjobscript,subdir+'/')
    os.chdir(subdir)

    if len(kwargs['param_files']) == 0:
        raise ValueError('param_files not given correctly in in.optzer')
    else:
        for pfile in kwargs['param_files']:
            if os.path.exists(pfile):
                os.remove(pfile)
        vars2params(varsfp['variables'], **kwargs)

    #...Do sub-jobs in the subdir_###
    L_up_lim = kwargs['fval_upper_limit']
    if print_level > 1:
        print('Performing subjobs at '+subdir, flush=True)
    try:
        prefix = kwargs['subjob-prefix']
        timeout= kwargs['subjob-timeout']
        cmd = prefix +" ./{0:s} > log.iid_{1:d}".format(subjobscript,kwargs['iid'])
        subprocess.run(cmd,shell=True,check=True,timeout=timeout)
        optdata = get_data('.',**kwargs)
        L = loss_func(optdata,**kwargs)
    except Exception as e:
        if print_level > 0:
            print('  Since subjobs failed at {0:s}, '.format(subdir)
                  +'the upper limit value is applied to its loss function.',
                  flush=True)
        os.chdir(cwd)
        L = L_up_lim

    try:
        #...Store data in iid_### directories
        os.mkdir("iid_{0:d}".format(kwargs['iid']))
        txt = ''
        for t in kwargs['target']:
            files = glob.glob(f'data*{t}')
            if f'data.ref.{t}' in files:
                files.remove(f'data.ref.{t}')
            txt += ' {0:s}'.format(files[0])
        for f in kwargs['param_files']:
            txt += f' {f}'
        for f in glob.glob('out.*'):
            txt += f' {f}'
        os.system("cp {0} iid_{1:d}/".format(txt,kwargs['iid']))
        os.chdir(cwd)
    except Exception as e:
        if print_level > 0:
            print(' Something wrong with storing data in iid_{0:d}'.format(kwargs['iid']))
            print(e)
        
        
    return L
    
def main():

    args = docopt(__doc__.format(os.path.basename(sys.argv[0])))
    headline()
    start = time.time()

    nproc = int(args['--nproc'])
    seed = int(args['--random-seed'])
    if seed < 0:
        seed = int(start)
        print(f' Random seed was set from the current time: {seed:d}')
    else:
        print(f' Random seed was given: {seed:d}')
    
    infp = read_in_optzer(_infname)
    write_info(infp,args)

    vs,vrs,vrsh,options,vopts = read_vars_optzer(infp['vars_file'])

    kwargs = infp
    kwargs['options'] = options
    kwargs['hardlim'] = vrsh
    # kwargs['infp'] = infp
    kwargs['subdir-prefix'] = args['--subdir-prefix']
    kwargs['subjob-script'] = args['--subjob-script']
    kwargs['subjob-prefix'] = args['--subjob-prefix']
    kwargs['subjob-timeout'] = int(args['--subjob-timeout'])
    kwargs['start'] = start
    kwargs['vopts'] = vopts
    
    if len(kwargs['target']) != 0:
        refdata = get_data('.',prefix='ref',**kwargs)
    else:
        raise ValueError(f'target may not given in {_infname}.')
    kwargs['refdata'] = refdata

    if len(kwargs['param_files']) == 0:
        raise ValueError(f'param_files should be given in {_infname}.')
    else:
        for fname in kwargs['param_files']:
            with open(fname,'r') as f:
                kwargs[fname] = f.read()

    print('\n # iid,losses=      iid',end='')
    if len(kwargs['target']) > 0:
        for t in kwargs['target']:
            print('  {0:>9s}'.format(t),end='')
    print('      total')

    maxiter = kwargs['num_iteration']
    if kwargs['opt_method'] in ('de','DE'):
        N = infp['de_num_individuals']
        F = infp['de_fraction']
        T = infp['de_temperature']
        CR = infp['de_crossover_rate']
        opt = DE(N,F,CR,T, vs,vrs,vrsh, func_wrapper, write_vars_optzer,
                 nproc=nproc, seed=seed, **kwargs)
    elif kwargs['opt_method'] in ('cs','CS','cuckoo','Cuckoo'):
        N = infp['cs_num_individuals']
        F = infp['cs_fraction']
        opt = CS(N,F, vs,vrs,vrsh, func_wrapper, write_vars_optzer,
                 nproc=nproc, seed=seed, **kwargs)
    elif kwargs['opt_method'] in ('tpe','TPE','wpe','WPE'):
        nbatch = nproc
        opt = TPE(nbatch, vs, vrs, vrsh, func_wrapper, write_vars_optzer,
                  seed=seed, **kwargs)
    
    opt.run(maxiter)

    print('\n optzer finished since it exceeds the max interation.')
    print(' Elapsed time = {0:.1f} sec.'.format(time.time()-start))
    
    return None

def headline():
    print('')
    print(' optzer --- optimize parameters of any external program to any target property ---')
    print('')
    cmd = ' '.join(s for s in sys.argv)
    print('   Executed as {0:s}'.format(cmd))
    hostname = subprocess.run(['hostname',], stdout=subprocess.PIPE).stdout.decode('utf-8')
    print('            at {0:s} on {1:s}'.format(os.getcwd(),hostname.strip()))
    print('            at {0:s}'.format(datetime.now().strftime('%Y-%m-%d %H:%M:%S')))
    print()
    print('   Please cite:')
    print('     1) R. Kobayashi, J. Open Source Software, 6(57), 2768 (2021)')
    print()
    return None

if __name__ == "__main__":

    main()



#!/usr/bin/env python
"""
Tree-based Parzen Estimator (TPE).

Usage:
  {0:s} [options]

Options:
  -h, --help  Show this message and exit.
"""
from __future__ import print_function

import os, sys
from docopt import docopt
import numpy as np
import random
import copy
from multiprocessing import Process, Pool
from time import time

__author__ = "RYO KOBAYASHI"
__version__ = "rev220123"

def test_func(var, vranges, **kwargs):
    x,y= var
    res= x**2 +y**2 +100.0*exp(-x**2 -y**2)*sin(2.0*(x+y))*cos(2*(x-y)) \
         +80.0*exp(-(x-1)**2 -(y-1)**2)*cos(x+4*y)*sin(2*x-y) \
         +200.0*sin(x+y)*exp(-(x-3)**2-(y-1)**2)
    return res

def test_write_func(vs,vrs,fname,**kwargs):
    with open(fname,'w') as f:
        for i,v in enumerate(vs):
            vr = vrs[i]
            f.write(' {0:10.3f}  {1:10.3f}  {2:10.3f}\n'.format(v,*vr))
    return None

def random_variables(vranges):
    """
    Create random variables within the given ranges.
    """
    rvs = np.empty(len(vranges))
    for i in range(len(vranges)):
        vmin, vmax = vranges[i]
        rvs[i] = np.random.random()*(vmax -vmin) +vmin
        # print(' i,vmin,vmax,v=',i,vmin,vmax,v)
    return rvs

def gauss_kernel(x):
    return np.exp(-0.5*x*x)/np.sqrt(2.0*np.pi)

class Sample:
    """
    A sample point that has parameter vector and loss value.
    """
    def __init__(self, iid, ndim, loss_func):
        self.iid = iid
        self.ndim = ndim
        self.loss_func = loss_func
        self.variables = np.zeros(self.ndim)
        self.val = None
        return None

    def set_variables(self,variables):
        if len(variables) != len(self.variables):
            raise ValueError('len(variables) != len(self.variables)')

        self.variables[:] = variables[:]
        self.val = None
        return None

    def calc_loss_func(self, vlogs, **kwargs):
        """
        Compute loss function value using self.loss_func function given in the constructor.
        """
        vec = copy.copy(self.variables)
        # If vlog != None, some of variables may be expressed in log domain,
        # they must be transformed back to non-log domain.
        for i in range(len(vec)):
            if vlogs[i]:
                vec[i] = np.exp(vec[i])
        val = self.loss_func(vec, **kwargs)
        return val,kwargs['index']

class TPE:
    """
    Class for Tree-based Parzen Estimator (TPE) or Weighted Parzen Estimator (WPE).
    
    Refs:
      1. Bergstra, J., Bardenet, R., Bengio, Y. & Kégl, B.  in Proc. NIPS-24th, 2546–2554
    """

    def __init__(self, nbatch, variables, vranges, vlimits, loss_func,
                 write_func, seed=42, **kwargs):
        """
        Conctructor of TPE class.

        Parameters:
          nbatch : int
            Number of samples in a batch which is the same as num of processes.
          variables : 1-D np.array
            np.array of variables to be optimized.
          vranges : 2-D np.array
            Current lower and upper limit of variables, which are used only at random sampling processes.
          vlimits : 2-D np.array
            Hard limits of variables, which are fixed during TPE iterations. 
          loss_func : function
            Loss function to be minimized with variables and **kwargs.
          write_func : function
            Function for outputing some info.
        """
        if nbatch < 1:
            raise ValueError('nbatch must be > 0.')
        np.random.seed(seed)
        random.seed(seed)
        self.nbatch = nbatch
        self.ndim = len(variables)
        self.vars0 = variables
        self.vranges = vranges
        self.vlimits = vlimits
        self.vlogs = [ False for i in range(len(self.vars0))]
        if 'vlogs' in kwargs.keys():
            self.vlogs = kwargs['vlogs']
        self.loss_func = loss_func
        self.write_func = write_func
        self.kwargs = kwargs
        self.best_pnt = None
        self.print_level = 0
        self.nsmpl_prior = 100
        self.ntrial = 100
        self.method = kwargs['opt_method']
        self.fname_smpl = 'out.{0:s}.samples'.format(self.method)

        self.gamma = 0.15
        #...Change default values if specified
        if 'print_level' in kwargs.keys():
            self.print_level = int(kwargs['print_level'])
        if 'tpe_nsmpl_prior' in kwargs.keys():
            self.nsmpl_prior = int(kwargs['tpe_nsmpl_prior'])
        if 'tpe_ntrial' in kwargs.keys():
            self.ntrial = int(kwargs['tpe_ntrial'])
        if 'tpe_gamma' in kwargs.keys():
            self.gamma = float(kwargs['tpe_gamma'])

        if self.gamma < 0.0 or self.gamma > 1.0:
            raise ValueError('gamma must be within 0. and 1., whereas gamma = ',self.gamma)

        #...Write info
        print('')
        if self.method in ('wpe','WPE'):
            print('   {0:s} infomation:'.format(self.method))
            print(f'     Num of prior samples = {self.nsmpl_prior:d}')
            print(f'     Num of top samples used for density estimation = {self.ntrial:d}')
        elif self.method in ('tpe','TPE'):
            print('   {0:s} infomation:'.format(self.method))
            print(f'     Num of prior samples = {self.nsmpl_prior:d}')
            print(f'     Num of trials for sampling = {self.ntrial:d}')
            print(f'     Gamma for dividing high and low = {self.gamma:4.2f}')
        print('')
    
        #...Change vrange if log domain
        for i in range(self.ndim):
            if self.vlogs[i]:
                self.vars0[i] = np.log(self.vars0[i])
                for l in range(2):
                    self.vranges[i,l] = np.log(self.vranges[i,l])
                    self.vlimits[i,l] = np.log(self.vlimits[i,l])

        #...Initialize sample history
        self.history = []  # History of all samples
        self.iidmax = 0
        for i in range(self.nbatch):
            self.iidmax += 1
            smpl = Sample(self.iidmax, self.ndim, self.loss_func)
            if i == 0:
                smpl.set_variables(self.vars0)
            else:
                smpl.set_variables(random_variables(self.vranges))
            self.history.append(smpl)

        return None

    def keep_best(self):
        vals = []
        for i,si in enumerate(self.history):
            if si.val == None:
                raise ValueError('Something went wrong.')
            vals.append(si.val)

        minval = min(vals)
        if self.bestsmpl == None or minval < self.bestsmpl.val:
            idx = vals.index(minval)
            self.bestsmpl = copy.deepcopy(self.history[idx])
        return None

    def run(self,maxstp=0):
        """
        Perfom TPE.
        """

        starttime = time()

        fsmpl = open(self.fname_smpl,'w')
        #...Headers
        fsmpl.write('# {0:>7s}  {1:>12s}'.format('iid', 'loss'))
        for i in range(len(self.history[0].variables)):
            fsmpl.write(' {0:>8d}-th'.format(i+1))
        fsmpl.write('\n')

        #...Create pool before going into maxstp-loop,
        #...since creating pool inside could cause "Too many files" error.
        pool = Pool(processes=self.nbatch)

        #...Evaluate sample losses of initial sets
        prcs = []
        for i,ci in enumerate(self.history):
            kwtmp = copy.copy(self.kwargs)
            kwtmp['index'] = i
            kwtmp['iid'] = ci.iid
            prcs.append(pool.apply_async(ci.calc_loss_func, (self.vlogs,), kwtmp))

        results = [ res.get() for res in prcs ]
        for res in results:
            val, i = res
            self.history[i].val = val
        #self.history.extend(candidates) # not need

        self._update_vranges()
        
        #...Check best
        self.bestsmpl = self.history[0]
        for si in self.history[1:]:
            if si.val < self.bestsmpl.val:
                self.bestsmpl = si
        fname = 'in.vars.optzer.{0:d}'.format(self.bestsmpl.iid)
        self.write_variables(self.bestsmpl,
                             fname=fname,
                             **self.kwargs)
        os.system('cp -f {0:s} in.vars.optzer.best'.format(fname))
        
        #...Write sample data
        for i,si in enumerate(self.history):
            self._write_smpl_data(fsmpl,si)

        if self.print_level > 0:
            self._write_step_info(0,starttime)

        #...TPE loop starts
        for istp in range(1,maxstp):
            #...Create candidates by either random or TPE
            if len(self.history) <= self.nsmpl_prior:
                #...Create random candidates
                candidates = []
                for i in range(self.nbatch):
                    self.iidmax += 1
                    newsmpl = Sample(self.iidmax, self.ndim, self.loss_func)
                    newsmpl.set_variables(random_variables(self.vranges))
                    candidates.append(newsmpl)
            else:
                if self.method in ('wpe,''WPE'):
                    #...Create candidates by WPE
                    candidates = self._candidates_by_WPE()
                else:
                    #...Create candidates by TPE
                    candidates = self._candidates_by_TPE()
                    

            #...Evaluate sample losses
            prcs = []
            for i,ci in enumerate(candidates):
                kwtmp = copy.copy(self.kwargs)
                kwtmp['index'] = i
                kwtmp['iid'] = ci.iid
                prcs.append(pool.apply_async(ci.calc_loss_func, (self.vlogs,), kwtmp))

            results = [ res.get() for res in prcs ]
            for res in results:
                val, i = res
                candidates[i].val = val
            self.history.extend(candidates)

            self._update_vranges()

            #...Check the best
            best_updated = False
            for si in self.history[-self.nbatch:]:
                if si.val < self.bestsmpl.val:
                    self.bestsmpl = si
                    best_updated = True
            if best_updated:
                fname = 'in.vars.optzer.{0:d}'.format(self.bestsmpl.iid)
                self.write_variables(self.bestsmpl,
                                     fname=fname,
                                     **self.kwargs)
                os.system('cp -f {0:s} in.vars.optzer.best'.format(fname))

            for i,si in enumerate(self.history[-self.nbatch:]):
                self._write_smpl_data(fsmpl,si)

            #...Write info
            if self.print_level > 0:
                self._write_step_info(istp,starttime)
            
        fsmpl.close()
        pool.close()
        return None

    def _write_smpl_data(self,f,smpl):
        f.write(' {0:8d}  {1:12.4e}'.format(smpl.iid, smpl.val))
        for j,vj in enumerate(smpl.variables):
            f.write(' {0:11.3e}'.format(vj))
        f.write('\n')
        return None

    def _write_step_info(self,istp,starttime):
        print(' step,time,best,vars='
              +' {0:6d} {1:8.1f}  {2:8.4f}'.format(istp+1, time()-starttime,
                                                   self.bestsmpl.val),end="")
        for i in range(min(16,self.ndim)):
            print(' {0:6.3f}'.format(self.bestsmpl.variables[i]),end="")
        print('', flush=True)
        return None

    def _update_vranges(self,):
        """
        Update variable ranges from the top rankers in the history.
        New vranges are obtained as max distances among top rankers from the best.
        """
        vals = np.zeros(len(self.history))
        for i,si in enumerate(self.history):
            vals[i] = si.val
        if len(self.history) > self.ntrial:
            iargs = np.argpartition(vals,self.ntrial)
            tmpsmpls = [ self.history[i] for i in iargs[:self.ntrial] ]
            vals = vals[ iargs[:self.ntrial] ]
        else:
            tmpsmpls = copy.copy(self.history)
        imin = np.argmin(vals)
        vmin = vals[imin]
        xtmps = np.zeros((len(tmpsmpls),self.ndim))
        for i in range(len(tmpsmpls)):
            xtmps[i,:] = tmpsmpls[i].variables[:]
        xmin = xtmps[imin,:]
        xwdth = np.zeros(self.ndim)
        for ismpl in range(len(xtmps)):
            if ismpl==imin: continue
            xi = xtmps[ismpl,:]
            w = np.abs(xi-xmin)
            for i in range(len(w)):
                xwdth[i] = max(w[i],xwdth[i])
        for idim in range(self.ndim):
            self.vranges[idim,0] = max(xmin[idim]-xwdth[idim],self.vlimits[idim,0])
            self.vranges[idim,1] = min(xmin[idim]+xwdth[idim],self.vlimits[idim,1])
        return None

    def _candidates_by_TPE(self,):
        """
        Create candidates by using TPE.
        """
        vals = np.zeros(len(self.history))
        for i,si in enumerate(self.history):
            vals[i] = si.val
        nlow = int(self.gamma *len(self.history))
        nhigh = len(self.history) -nlow
        argpart = np.argpartition(vals,nlow)
        Xlow = np.zeros((nlow,self.ndim))
        Xhigh = np.zeros((nhigh,self.ndim))
        for i in range(nlow):
            iid = argpart[i]
            Xlow[i,:] = self.history[iid].variables[:]
        for i in range(nhigh):
            iid = argpart[nlow+i]
            Xhigh[i,:] = self.history[iid].variables[:]

        #...Sampling variable candidates
        xcandidates = np.empty((self.nbatch,self.ndim))
        ntrial = max(self.ntrial,self.nbatch)
        for idim in range(self.ndim):
            xhmin = self.vlimits[idim,0]
            xhmax = self.vlimits[idim,1]
            xlowsrt = np.sort(Xlow[:,idim])
            npnt = len(xlowsrt)
            #...Determine smoothness parameter, h, by Silverman's method
            q75, q25 = np.percentile(xlowsrt, [75,25])
            std = np.std(xlowsrt)
            sgm = min(std, (q75-q25)/1.34)
            h = 1.06 *sgm /np.power(npnt, 1.0/5)
            #...Prepare for g(x)
            xhighsrt = Xhigh[:,idim]
            q75, q25 = np.percentile(xhighsrt, [75,25])
            sgmh = min(np.std(xhighsrt), (q75-q25)/1.34)
            hh = 1.06 *sgmh /np.power(len(xhighsrt),1.0/5)
            #...Several trials for selection
            aquisition = np.zeros(ntrial)
            xs = np.empty(ntrial)
            for itry in range(ntrial):
                ipnt = int(np.random.random()*npnt)
                xi = xlowsrt[ipnt]
                r = h *np.sqrt(-2.0*np.log(np.random.random()))
                th = 2.0 *np.pi *np.random.random()
                x = xi + r*np.cos(th)
                #...Wrap by vranges
                x = min(max(x,xhmin),xhmax)
                xs[itry] = x
                #...Compute l(x) and g(x)
                lx = 0.0
                for j in range(npnt):
                    z = (x-xlowsrt[j])/h
                    lx += np.exp(-0.5*z*z)
                lx /= npnt*h *np.sqrt(2.0*np.pi)
                gx = 0.0
                for j in range(len(xhighsrt)):
                    z = (x-xhighsrt[j])/hh
                    gx += np.exp(-0.5*z*z)
                gx /= len(xhighsrt)*hh *np.sqrt(2.0*np.pi)
                aquisition[itry] = gx/lx
            #...Pick nbatch of minimum aquisition points
            idxsort = np.argsort(aquisition)
            xcandidates[:,idim] = xs[idxsort[0:self.nbatch]]
        #...Create sample with xcandidate as variables
        candidates = []
        for ib in range(self.nbatch):
            self.iidmax += 1
            smpl = Sample(self.iidmax, self.ndim, self.loss_func)
            smpl.set_variables(xcandidates[ib,:])
            candidates.append(smpl)
        return candidates

    def _candidates_by_WPE(self,):
        """
        Create candidates by using WPE.
        """
        vals = np.zeros(len(self.history))
        for i,si in enumerate(self.history):
            vals[i] = si.val
        if len(self.history) > self.ntrial:
            iargs = np.argpartition(vals,self.ntrial)
            tmpsmpls = [ self.history[i] for i in iargs[:self.ntrial] ]
            vals = vals[ iargs[:self.ntrial] ]
        else:
            tmpsmpls = copy.copy(self.history)
        vmin = vals.min()
        wgts = np.array([ np.exp(-(v-vmin)/vmin) for v in vals ])
        xtmps = np.zeros((len(tmpsmpls),self.ndim))
        for i in range(len(tmpsmpls)):
            xtmps[i,:] = tmpsmpls[i].variables[:]
        #...Sampling variable candidates
        xcandidates = np.empty((self.nbatch,self.ndim))
        for idim in range(self.ndim):
            xhmin = self.vlimits[idim,0]
            xhmax = self.vlimits[idim,1]
            xsrt = np.sort(xtmps[:,idim])
            #...Determine smoothness parameter, h, by Silverman's method
            q75, q25 = np.percentile(xsrt, [75,25])
            sgm = min(np.std(xsrt), (q75-q25)/1.34)
            h = 1.06 *sgm /np.power(len(xsrt), 1.0/5)
            xs = np.empty(self.nbatch)
            for ib in range(self.nbatch):
                ipnt = random.choices([j for j in range(len(wgts))],weights=wgts)[0]
                xi = xtmps[ipnt,idim]
                r = h *np.sqrt(-2.0*np.log(np.random.random()))
                th = 2.0 *np.pi *np.random.random()
                x = xi + r*np.cos(th)
                #...Wrap by vranges
                x = min(max(x,xhmin),xhmax)
                xs[ib] = x
            xcandidates[:,idim] = xs[:]
        
        #...Create sample with xcandidate as variables
        candidates = []
        for ib in range(self.nbatch):
            self.iidmax += 1
            smpl = Sample(self.iidmax, self.ndim, self.loss_func)
            smpl.set_variables(xcandidates[ib,:])
            candidates.append(smpl)
        
        return candidates

    def write_variables(self,smpl,fname='in.vars.optzer',**kwargs):
        vs = smpl.variables
        vrs = self.vranges
        self.write_func(vs,vrs,fname,**kwargs)
        return None

def main():
    args = docopt(__doc__.format(os.path.basename(sys.argv[0])))
    return None

if __name__ == "__main__":

    main()



import os
import numpy as np

__author__ = "RYO KOBAYASHI"
__version__ = "221223"

def read_in_optzer(fname='in.optzer'):
    #...initialize
    infp = {}
    infp['fval_upper_limit'] = 100.0
    infp['missing_value'] = 1.0
    infp['print_level'] = 1
    infp['weights'] = {'rdf':1.0, 'adf':1.0, 'vol':1.0, 'lat':1.0}
    infp['update_vrange'] = -1
    infp['vars_file'] = 'in.vars.optzer'

    mode = None
    infp['target'] = []
    infp['param_files'] = []
    
    with open(fname,'r') as f:
        lines = f.readlines()

    for line in lines:
        if line[0] in ('!','#'):
            mode = None
            continue
        data = line.split()
        if len(data) == 0:
            mode = None
            continue
        if data[0] == 'num_iteration':
            maxiter = int(data[1])
            infp['num_iteration'] = maxiter
            mode = None
        elif data[0] == 'print_level':
            print_level = int(data[1])
            infp['print_level'] = print_level
            mode = None
        elif data[0] == 'opt_method':
            opt_method = data[1]
            infp['opt_method'] = opt_method
            mode = None
        elif data[0] == 'param_files':
            infp[data[0]] = [ name for name in data[1:] ]
            mode = None
        elif data[0] == 'fval_upper_limit':
            fup_limit = float(data[1])
            infp['fval_upper_limit'] = fup_limit
            mode = None
        elif data[0] == 'missing_value':
            misval = float(data[1])
            infp['missing_value'] = misval
            mode = None
        elif data[0] in ('target'):
            if len(data) < 2:
                raise RuntimeError('target entry requires at least one keyword.')
            for i in range(1,len(data)):
                infp['target'].append(data[i])
            mode = None
        elif data[0] == 'de_num_individuals':
            nind = int(data[1])
            infp['de_num_individuals'] = nind
            mode = None
        elif data[0] == 'de_crossover_rate':
            cr = float(data[1])
            infp['de_crossover_rate'] = cr
            mode = None
        elif data[0] == 'de_fraction':
            frac = float(data[1])
            infp['de_fraction'] = frac
            mode = None
        elif data[0] == 'de_temperature':
            temp = float(data[1])
            infp['de_temperature'] = temp
            mode = None
        elif data[0] == 'cs_num_individuals':
            nind = int(data[1])
            infp['cs_num_individuals'] = nind
            mode = None
        elif data[0] == 'cs_fraction':
            frac = float(data[1])
            infp['cs_fraction'] = frac
            mode = None
        elif data[0] == 'tpe_gamma':
            infp['tpe_gamma'] = float(data[1])
            mode = None
        elif '_nsmpl_prior' in data[0]:
            infp['tpe_nsmpl_prior'] = int(data[1])
            mode = None
        elif '_ntrial' in data[0]:
            infp['tpe_ntrial'] = int(data[1])
            mode = None
        elif data[0] == 'update_vrange':
            infp['update_vrange'] = int(data[1])
            mode = None
        else:
            mode = None
            pass
    
    return infp

def read_out_optzer(fname='out.optzer'):
    with open(fname,'r') as f:
        lines = f.readlines()
    bestiid = -1
    for line in reversed(lines):
        if 'step,time' in line:
            data = line.split()
            bestiid = int(data[3])
            bestloss = float(data[4])
            break
    if bestiid < 0:
        raise ValueError(f'Failed to get best_iid from {fname}!!!')

    targets = []
    weights = []
    losses = []
    mode = ''
    for line in lines:
        if 'weights:' in line:
            mode = 'weights'
            continue
        if '# iid,losses' in line:
            data = line.split()
            targets = [ t for t in data[3:] ]
            continue
        elif 'iid,losses' in line and ' {0:d} '.format(bestiid) in line:
            if len(targets) < 1:
                raise ValueError('len(targets) < 1 !!!')
            data = line.split()
            losses = [ float(l) for l in data[2:] ]
            break
        if mode == 'weights':
            data = line.split()
            if len(data) != 2:
                mode = ''
                continue
            weights.append(float(data[1]))

    return bestiid,targets,weights,losses

def write_info(infp,args):
    """
    Write out information on input parameters for fp.
    """

    print('\n Input')
    print(' ----------')
    print('   num of processes (given by --nproc option)  ',int(args['--nproc']))
    try:
        if len(infp['param_files']) == 0:
            print('   potential       {0:s}'.format(infp['potential']))
        else:
            print('   param_files  ',end='')
            for fname in infp['param_files']:
                print(f'  {fname}', end='')
            print('')
    except:
        raise

    fmethod = infp['opt_method']
    print('   opt_method  {0:s}'.format(fmethod))
    if fmethod in ('de','DE'):
        print('   num_individuals   ',infp['de_num_individuals'])
        print('   fraction          {0:7.4f}'.format(infp['de_fraction']))
        print('   temparature       {0:7.4f}'.format(infp['de_temperature']))
        print('   crossover_rate    {0:7.4f}'.format(infp['de_crossover_rate']))
    elif fmethod in ('cs','CS'):
        print('   num_individuals   ',infp['cs_num_individuals'])
        print('   fraction          {0:7.4f}'.format(infp['cs_fraction']))
    elif fmethod in ('tpe','TPE','wpe','WPE'):
        pass
    else:
        print('   There is no such fitting method...')
    print('   num_iteration   {0:d}'.format(infp['num_iteration']))
    print('   missing_value   {0:.1f}'.format(infp['missing_value']))
    print(' ----------')
    print()
    return None

def write_vars_optzer(vs,vrs,fname='in.vars.optzer',**kwargs):
    options = kwargs['options']
    hardlim = kwargs['hardlim']
    vopts = kwargs['vopts']
    nv = len(vs)
    with open(fname,'w') as f:
        if 'hard-limit' in options.keys() and options['hard-limit']:
            f.write('!  hard-limit:  T\n')
            f.write('!\n')
        f.write(' {0:5d} \n'.format(nv))
        for i in range(len(vs)):
            f.write(' {0:15.7f}  {1:15.7f}  {2:15.7f}'.format(vs[i],*vrs[i])
                    +'  {0:10.4f}  {1:10.4f}'.format(*hardlim[i]))
            if len(vopts[i]) > 0:
                for ivo,vo in enumerate(vopts[i]):
                    f.write(f'  {vo}')
            f.write('\n')
    return None

def read_vars_optzer(fname='in.vars.optzer'):
    with open(fname,'r') as f:
        lines = f.readlines()
    iv = 0
    nv = -1
    vs = []
    vrs = []
    vrsh = []
    vopts = []
    options = {}
    for line in lines:
        if line[0] in ('!','#'):
            k,v = parse_option(line)
            if k is not None:
                options[k] = v[0]
                #print(' option: ',k,v)
            continue
        data = line.split()
        if len(data) == 0:
            continue
        if nv < 0:
            nv = int(data[0])
            continue
        else:
            iv += 1
            if iv > nv:
                break
            vs.append(float(data[0]))
            vrs.append([ float(data[1]), float(data[2])])
            vrsh.append([float(data[3]), float(data[4])])
            vopts.append(data[5:])
            # print(' iv,vrhmin,vrhmax= {0:3d} {1:11.3e} {2:11.3e}'.format(iv,
            #                                                              float(data[3]),
            #                                                              float(data[4])))
    vs = np.array(vs)
    vrs = np.array(vrs)
    vrsh = np.array(vrsh)
    print('')
    return vs,vrs,vrsh,options,vopts
    

def read_data(fname,):
    """
    General routine of reading data.
    
    Input file format
    -----------------
    ```
    #  Comment lines begins with '#' or '!'
    #  Options start with "option-name: "
    10    1.0
    0.1234  0.2345  0.3456  0.4567  0.5678  0.6789
    0.7890  0.8901  0.9012  0.0123
    ```
    - 1st line:  num of data (NDAT),  weight of the data (WDAT)
    - 2nd line-: data values (number of data should be equal to NDAT)
    ----------------------------------------
    In case that "datatype: independent" is specified in the option,
    the input file format is changed as the following,
    ```
      10     1.0
      0.1234    0.1234
     -0.2345    0.2345
      0.3456    0.1
      ...
    ```
    - Below the 1st line-:  (value, error eps) pair
    """
    if not os.path.exists(fname):
        raise RuntimeError('File not exsits: ',fname)
    
    with open(fname,'r') as f:
        lines = f.readlines()

    ndat = 0
    wdat = 0.0
    data = None
    epss = None
    idat = 0
    done = False
    options = {'datatype': 'continuous', 'eps':1.0e-3}
    for line in lines:
        if line[0] in ('#','!'):
            try: 
                k,v = parse_option(line)
            except:
                k = None
                v = None
            if k != None and v != None: # valid option
                if len(v) == 1: # options that take only one argument
                    options[k] = v[0]
                else: # options that take more than one arguments
                    if k == 'subject_to':
                        if len(v) != 4:
                            raise ValueError('Num of arguments for subject_to option is wrong, len(v)= ',len(v))
                        if 'subject_to' not in options.keys():
                            options[k] = []
                        options[k].append({'pid':int(v[0]),
                                           'lower':v[1],
                                           'upper':v[2],
                                           'penalty':float(v[3])})
                    else:
                        options[k] = v
            continue
        ldat = line.split()
        if ndat < 1:
            ndat = int(ldat[0])
            wdat = float(ldat[1])
            data = np.zeros(ndat)
            if 'indep' in options['datatype']:
                epss = np.zeros(ndat)
        else:
            if data is None:
                raise RuntimeError('data is None, which should not happen.')
            if 'indep' in options['datatype']:
                # In case of datatype==independent, each line has (value,eps) pair information
                data[idat] = float(ldat[0])
                epss[idat] = float(ldat[1])
                idat += 1
                if idat == ndat:
                    done = True
                    break
            else:
                for i,d in enumerate(ldat):
                    data[idat] = float(d)
                    idat += 1
                    if idat == ndat:
                        done = True
                        break
        if done:
            break
    options['eps'] = float(options['eps'])
    if 'indep' in options['datatype']:
        options['epss'] = epss
    return {'ndat':ndat, 'wdat':wdat, 'data':data, **options}

def parse_option(line):
    """
    Parse option from a comment line.
    """
    words = line.split()
    if len(words) < 2 or words[1][-1] != ':':
        return None,None
    optname = words[1]
    k = words[1][:-1]
    v = words[2:]
    return k,v

def read_out_cs_gen(fname='out.cs.generations'):
    """Read out.cs.generations file."""
    with open(fname,'r') as f:
        lines = f.readlines()
    gens = []
    iids = []
    losses = []
    for l in lines:
        if l[0] == '#':
            continue
        data = l.split()
        gens.append(int(data[0]))
        iids.append(int(data[1]))
        losses.append(float(data[2]))
    return gens,iids,losses

def read_out_cs_ind(fname='out.cs.individuals'):
    """Read out.cs.individuals file."""
    with open(fname,'r') as f:
        lines = f.readlines()
    iids = []
    losses = []
    prms = []
    for l in lines:
        if l[0] == '#':
            continue
        data = l.split()
        iids.append(int(data[0]))
        losses.append(float(data[1]))
        prms.append([ float(d) for d in data[2:]])
    return iids,losses,prms
